package com.mckessonaps.jspwiki.ntauth;

import java.util.Properties;

import jcifs.UniAddress;
import jcifs.smb.NtlmPasswordAuthentication;
import jcifs.smb.SmbAuthException;
import jcifs.smb.SmbSession;

import org.apache.log4j.Logger;

import com.ecyrd.jspwiki.auth.UserProfile;
import com.ecyrd.jspwiki.auth.WikiAuthenticator;
import com.ecyrd.jspwiki.providers.ProviderException;

/**
 * Provides an NT based authentication against a domain controller. This module relies on the jcifs package
 * which can be found at: http://jcifs.samba.org/
 * 
 * @author Howard Henson
 * @author John Volkar
 */
public class NTAuthenticator implements WikiAuthenticator
{
    private static final Logger log = Logger.getLogger(NTAuthenticator.class);

    private String m_domainName;
    private String m_domainController;

    /**
     * Authenticates against an NT server using the jcifs package.
     */
    public boolean authenticate(UserProfile userProfile)
    {
        String userName = userProfile.getName();
        String userPass = userProfile.getPassword();

        if ((null == userName) || (0 == userName.length()))
            return false;

        boolean success = false;
        try
        {
            UniAddress domainController = UniAddress.getByName(m_domainController);

            NtlmPasswordAuthentication credentials =
                new NtlmPasswordAuthentication(m_domainName, userName, userPass);

            SmbSession.logon(domainController, credentials);

            success = true;
        }
        catch (SmbAuthException sae)
        {
            //ignore as user may just have mistyped his name or pass...
        }
        catch (Exception e)
        {
            log.warn("Exception caught " + this.toString(), e);
        }

        return success;
    }

    /**
     * Allow visibility of domain controller and name.
     */
    public String toString()
    {
        return "NTAuthenticator(domainController:'"
            + m_domainController
            + "'  domainName:'"
            + m_domainName
            + "')";
    }

    /**
     * Get the domain controller and domain name to be authenticated against.
     */
    public void initialize(Properties properties)
    {
        m_domainName = properties.getProperty("ntauthenticator.domain.name", "");
        m_domainController = properties.getProperty("ntauthenticator.domain.controller", "");
    }

    /**
     * Returns <code>false</code>.
     * 
     * @see WikiAuthenticator#canChangePasswords()
     */
    public boolean canChangePasswords()
    {
        return false;
    }

    /**
     * Not supported.
     * 
     * @see WikiAuthenticator#setPassword(UserProfile, String)
     */
    public void setPassword(UserProfile userProfile, String password) throws ProviderException
    {
        throw new ProviderException("NTAuthenticator does not support password changes.");
    }

}
