package com.mckessonaps.jspwiki.plugin.agregate;

import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;

import com.ecyrd.jspwiki.WikiContext;
import com.ecyrd.jspwiki.WikiPage;
import com.ecyrd.jspwiki.plugin.InsertPage;
import com.ecyrd.jspwiki.plugin.WikiPlugin;
import com.mckessonaps.jspwiki.plugin.util.ParameterException;

/**
 * Combine the QueryPlugin and InsertPage plugin to generate "aggregates" of the 
 * selected pages.  This could be used to auto-generate glossaries, dictionaries, 
 * and summaries of all sorts.
 * 
 * @author JVolkar  
 */
public class AggregatePlugin implements WikiPlugin
{

    private static final Comparator ALPHA = new AlphaPageComparator();
    private static final Comparator REVERSE_ALPHA = new ReverseAlphaPageComparator();
    private static final Comparator ASCENDING_MODIFICATION = new AscendingModificationPageComparator();
    private static final Comparator DESCENDING_MODIFICATION = new DescendingModificationPageComparator();

    private final AggregateParamGroup m_paramGroup = new AggregateParamGroup();

    
    //This gets newed for each instance of this plugin, but reused for each page
    //of the aggregation...
    private InsertPage m_insertPagePlugin = new InsertPage();

    /**
     * @see WikiPlugin#execute(WikiContext, Map)
     */
    public String execute(WikiContext context, Map paramMap)
    {
        try
        {
            m_paramGroup.update(paramMap);
        }
        catch (ParameterException e)
        {
            return e.getMessage();
        }

        return execute(context);
    }

    /**
     * Do the actual work of the aggregation.
     * 
     * @param context
     * @return the HTML output.
     */
    private String execute(WikiContext context)
    {
        StringBuffer buffer = new StringBuffer();

        try
        {
            Set querySet = executeQuery(context);
            List orderedList = applyOrdering(querySet, context);
            List aggregateList = limitMaxPages(orderedList);
            
            
            generateIndex(buffer, aggregateList, context);

            processAggregates(buffer, aggregateList, context);
        }
        catch (ParameterException pe)
        {
            buffer.setLength(0);
            buffer.append(pe.getMessage());
        }
        catch (AggregatePluginException ape)
        {
            buffer.setLength(0);
            buffer.append(ape.getMessage());
        }

        return buffer.toString();
    }

    /**
     * @param querySet
     * @return A sorted list of the elements in the querySet.
     */
    private List applyOrdering(Set querySet, WikiContext context) throws ParameterException
    {
        List sortedList = new LinkedList(querySet);

        namesToPages(sortedList, context);
        
        Comparator comp = getComparitor();

        Collections.sort(sortedList, comp);

        return sortedList;
    }

    /**
     * Transform a list of page names to a list of real WikiPage objects.
     * @param pageNameList
     */
    private void namesToPages(List pageNameList, WikiContext context)
    {
        for (int i = 0; i < pageNameList.size(); i++)
        {
            String pageName = (String)pageNameList.get( i );
            WikiPage page = context.getEngine().getPage( pageName );
            pageNameList.set( i, page );
        }
    }

    
    
    /**
     * @return The Comparitor to be used to sort the pages.
     */
    private Comparator getComparitor() throws ParameterException
    {
        String ordering = m_paramGroup.m_ordering.get();

        if (ordering.equalsIgnoreCase("alpha"))
            return ALPHA;

        if (ordering.equalsIgnoreCase("reverseAlpha"))
            return REVERSE_ALPHA;

        if (ordering.equalsIgnoreCase("ascendingModification"))
            return ASCENDING_MODIFICATION;

        if (ordering.equalsIgnoreCase("decendingModification"))
            return DESCENDING_MODIFICATION;

        m_paramGroup.m_ordering.throwBad( "Must be one of 'alpha', 'reverseAlpha', 'ascendingModification' or 'decendingModification'." );
        return null;
        
    }

    /**
     * @param orderedList
     * @return
     */
    private List limitMaxPages(List orderedList) throws ParameterException
    {
        int limit = m_paramGroup.m_maxPages.get();

        if (orderedList.size() <= limit)
            return orderedList;

        if (limit < 1) //Did user give a negative limit?
            m_paramGroup.m_maxPages.throwBad( "Must be >= '1'." );

        return orderedList.subList(0, limit);
    }

    /**
     * Create (or not) an index of the pages to be aggregated.
     * 
     * @param buffer
     * @param aggregateList
     * @param context
     */
    private void generateIndex(StringBuffer buffer, List aggregateList, WikiContext context) 
    throws ParameterException
    {
        String index = m_paramGroup.m_indexKind.get();
        
        if (index.equalsIgnoreCase("none"))
            return;

        String indexDelimiter = "";

        if (index.equalsIgnoreCase("delim"))
            indexDelimiter = ", ";
        else if (index.equalsIgnoreCase("list"))
            indexDelimiter = "<BR />";
        else 
            m_paramGroup.m_indexKind.throwBad( "Must be one of 'none', 'list' or 'delim'." );

        buffer.append("<DIV class=\"aggregationIndex\">");
        for (Iterator iter = aggregateList.iterator(); iter.hasNext();)
        {
            WikiPage page = (WikiPage) iter.next();
            buffer.append( buildAggregatedPageLink( page, context) );
            buffer.append(indexDelimiter);
        }
        buffer.setLength(buffer.length() - indexDelimiter.length()); //drop off the extra delim
        buffer.append("<BR />"); // always add a line break
        buffer.append("</DIV>");
    }

    /**
     * Forms an html anchor scrap to link to the given page.
     * @param page
     * @param context
     */
    private String buildAggregatedPageLink(WikiPage page, WikiContext context)
    {
        String indexLink = context.getEngine().getViewURL( page.getName() );
        String indexEntry = "<A href=\"" + indexLink +"\">" +page.getName() + "</A>";
        return  indexEntry ;
    }

    /**
     * Process each page in the aggregateList and output the data to the buffer.
     * 
     * @param buffer
     * @param aggregateList
     */
    private void processAggregates(StringBuffer buffer, List aggregateList, WikiContext context)
    {
        buffer.append("<DIV class=\"aggregation\">");

        String backToHtml = makeBackToHtml(context);
        String separatorHtml = makeSeparatorHtml();

        for (Iterator iter = aggregateList.iterator(); iter.hasNext();)
        {
            WikiPage page = (WikiPage) iter.next();

            buffer.append("<DIV class=\"aggregatedPage\">");
            
            buffer.append("<DIV class=\"aggregatedPageName\">");
            buffer.append( buildAggregatedPageLink( page, context) );
            buffer.append("<BR />");
            buffer.append("</DIV>"); //page name

            buffer.append(executeInsertPage(page, context));

            buffer.append(backToHtml);

            buffer.append(separatorHtml);
            buffer.append("</DIV>"); //page
        }

        buffer.append("</DIV>");
    }

    /**
     * @return The HTML needed for the back to feature.
     */
    private String makeBackToHtml(WikiContext context)
    {
        String backTo = m_paramGroup.m_backTo.get();
        
        if (backTo.equals(""))
            return "";

        String backTarget = "#section-" + context.getPage().getName() + "-" + backTo;

        backTo = "<DIV class=\"aggregatedBackTo\">";
        backTo += "<A href=\"" + backTarget + "\">Back</A>";
        backTo += "<BR />";
        backTo += "</DIV>";

        return backTo;
    }

    /**
     * @return The HTML to be user as the separtor between aggregated pages.
     */
    private String makeSeparatorHtml()
    {
        if (!m_paramGroup.m_emitSeparator.get())
            return "";

        String separatorHtml;
        separatorHtml = "<DIV class=\"aggregatedSeparator\">";
        separatorHtml += "<HR />";
        separatorHtml += "</DIV>";

        return separatorHtml;
    }

    /**
     * Execute the InsertPage plugin to get the data to be aggregated.
     * 
     * @param page
     */
    private String executeInsertPage(WikiPage page, WikiContext context)
    {
        Map paramMap = new HashMap();
        paramMap.put("section", Integer.toString(m_paramGroup.m_section.get()));
        paramMap.put("page", page.getName());
        paramMap.put("class", m_paramGroup.m_bodyClass.get());
        paramMap.put("maxlen", Integer.toString(m_paramGroup.m_maxlength.get()));

        return m_insertPagePlugin.execute(context, paramMap);
    }

    /**
     * Execute the Query needed to generate the set of pages to be aggregated...
     * 
     * @param context To be passed to the QueryPlugin for it's invokation.
     * @return The Set of pages to be aggregated.
     * @throws AggregatePluginException If anyhting goes wrong with the QueryPlugin invokation.
     */
    private Set executeQuery(WikiContext context) throws AggregatePluginException
    {
        String paramName = m_paramGroup.m_pages.getName();
        String paramValue = m_paramGroup.m_pages.get();
        
        Map paramMap = new HashMap();

        if (paramValue.equals(""))
            paramMap.put(paramName, "TO [" + context.getPage().getName() + "]");
        else
            paramMap.put(paramName, paramValue);

        QueryDelegate queryDelegate = new QueryDelegate();

        String result = queryDelegate.execute(context, paramMap);

        //Check for an error from the QueryPlugin, wrap and throw...
        if (!result.equals(queryDelegate.toString()))
            throw new AggregatePluginException(result);

        return queryDelegate.getResultSet();
    }

}
