package com.mckessonaps.jspwiki.plugin.emailnotifier;

import java.util.HashMap;
import java.util.Map;

import org.apache.log4j.Category;

import com.ecyrd.jspwiki.WikiContext;
import com.ecyrd.jspwiki.WikiEngine;
import com.ecyrd.jspwiki.WikiPage;
import com.ecyrd.jspwiki.plugin.WikiPlugin;
import com.ecyrd.jspwiki.providers.WikiPageProvider;

/**
 * Assure that for each wiki that we see a plugin execute request from that we have a
 * NotifierThread running. On each plugin execution, update the NotifierThread with the "current"
 * parameters to the plugin. This plugin simply returns a status message about when the notifier
 * will next run. <br/><b>NOTE:</b> If two instances of this plugin are placed in a wiki,
 * (either on the smae or differing pages), there is not harm to it, but the results will be
 * unpredictable in that the notification task will use the last set of parameters that it
 * recieved. Since parameters are updated each time the plugin is executed this menas that the last
 * viewed page with this plugin is the one whose parameters will get used. <br/><b>Usage: Simple
 * text message, to be built every hour at 5 minutes after the hour.</b>
 * 
 * <pre>
 *  [{INSERT com.mckessonaps.jspwiki.plugin.emailnotifier.EmailNotifierPlugin WHERE server='your-mail-server' from='your-wiki' subject='Your Wiki Changes' subscriptionPage='EmailNotifierSubscriptions' minuteOffset='5' hourInterval='1' changedPagePattern='* CHANGED_PAGE_NAME by CHANGED_PAGE_INFO\n'
 * 
 *  Hello, here are the changes in the last 24 hours. ----- CHANGED_PAGE_LIST ----- To unsubscribe see the SUBSCRIPTION_PAGE. }]
 * </pre>
 * 
 * <hr/>
 * <b>Design notes:</b><br/>Rough class hierarchy... <br/>Plugin --1> Map --*> Thread --1>
 * Task --*> Subscribers
 * 
 * 
 * @author JVolkar
 * @author StevenPozarycki
 */
public class EmailNotifierPlugin implements WikiPlugin
{
    private static final Category log = Category.getInstance(EmailNotifierPlugin.class);

    private static final Map c_notifierThreadMap = new HashMap();

    /**
     * Get the NotifierTask instance appropriate for the <code>context</code>, update it's
     * parameters, return it's status message.
     */
    public String execute(WikiContext context, Map paramMap)
    {
        NotifierThread thread = getNotifierThread(context);

        WikiPage requestedPage = context.getPage();
        WikiPage newestPage = context.getEngine().getPage(requestedPage.getName(), WikiPageProvider.LATEST_VERSION);

        //This sometimes happens, not sure what this means in a wiki, ignore it...
        if ((requestedPage != null) && (newestPage != null))
        {
            //Only update the parameters if it's the newest version of the page, if someone
            //is browsing the history of the page don't use the old parameters...
            if (requestedPage.getVersion() == newestPage.getVersion())
                thread.updateParameters(paramMap);
        }

        return thread.getNotifierTask().getStatusMessage();
    }

    /**
     * Get the correct task for the given context, if no task, create one.
     * 
     * @param context
     * @return NotifierThread
     */
    private NotifierThread getNotifierThread(WikiContext context)
    {
        WikiEngine engine = context.getEngine();
        String appName = engine.getApplicationName();

        NotifierThread thread = (NotifierThread) c_notifierThreadMap.get(appName);
        if (null == thread)
        {
            synchronized (c_notifierThreadMap)
            {
                //Try to get task again, someone could have created one before our first check and
                // acquiring the lock.
                thread = (NotifierThread) c_notifierThreadMap.get(appName);
                if (null == thread)
                {
                    thread = new NotifierThread(engine);
                    //Janne? what is the relation ship between WikiEngine instances and wiki
                    // installations? one to one? right?
                    thread.start();
                    c_notifierThreadMap.put(appName, thread);
                    log.info("new NotifierThread(): " + thread);
                }
            }
        }
        return thread;
    }

}
