package com.mckessonaps.jspwiki.plugin.emailnotifier;

import java.util.Map;

import org.apache.log4j.Category;

import com.ecyrd.jspwiki.WikiEngine;


/**
 * One instance of this thread is created per {@link WikiEngine#getApplicationName()}.  
 * This thread manages one NotifierTask which it periodically runs.  This thread is 
 * marked daemon and just spends most of it's time sleeping.
 * 
 * <hr/>
 * <b>Design Notes:</b><br/>
 * At one point early on in development this plugin used a Timer and TimerTasks.  Unfortunately
 * you cannot RE-scheudle a TimerTask, you can schedule it as a one-shot or periodic, and you 
 * can cancel it, but not reschedule or even cancel and schedule again.  This means that this 
 * plugin could not have it's schduling changed once started.  Oh well, we'll just brute-force 
 * it with this sleeping thread. 
 */
class NotifierThread 
extends Thread
{
    private static final Category log = Category.getInstance( NotifierThread.class );
    
    /** Since plugin's are executed every time a page is generated, we will see
     * alot of bogus updates of our parameters.  That's okay to a point, but wasteful
     * since each update causes this thread to be interrupted.  So we will at most 
     * accept new parameter updates every so often.  This is the minimum amount of 
     * time that must pass before we will accept a parameter update.*/
    static long MINIMUM_UPDATE_INTERVAL = 60000;
    
    private long m_lastParamUpdate = System.currentTimeMillis() - (MINIMUM_UPDATE_INTERVAL + 1); //just a little in the past...

    private final NotifierTask m_notifierTask;
    


    /**
     * Constructor, creates the task, sets this thread as deamon gives a name 
     * and logs thread construction.
     */
    NotifierThread(WikiEngine engine)
    {
        m_notifierTask = new NotifierTask( engine );
        setDaemon( true );
        setName( "NotifierThread for: " + engine.getApplicationName() );
        log.info( "new NotifierThread() "+ toString() );
    }
    
    
    /**
     * Returns reference to the task object.
     */
    NotifierTask getNotifierTask()
    {
        return m_notifierTask;
    }
    
    
    /**
     * Never ending main loop, idles and executes the task if it's time.
     * @see java.lang.Runnable#run()
     */
    public void run()
    {
        while (true)
        {
            idleUntilNotificationTime();
            
            if (isTimeToNotify())
                m_notifierTask.execute();
                
        }
    }
    
    
    /**
     * Pass the params onto the task so it can read them, interrupt this thread so 
     * it can reschedule itsself for the new params.
     */
    void updateParameters(Map paramMap)
    {
        //Hmmm, if the plugin is stuck on something like Main or LeftMenu or LeftMenuFooter this method 
        //will get called _a lot_.  Don't hammer too hard on this poor little method, only _really_ update 
        //the parameters at most once per MINIMUM_UPDATE_INTERVAL...
        long nowMillis = System.currentTimeMillis(); 
        if ((nowMillis - m_lastParamUpdate ) < MINIMUM_UPDATE_INTERVAL)
            return;
            
        m_lastParamUpdate = nowMillis;
        
        log.debug( "updateParameters()" );
        m_notifierTask.updateParameters( paramMap );
        
        interrupt(); 
    }
    
    
    /**
     * Computes how long to sleep, sleeps in a try catch block ignores interruptions.
     */
    private void idleUntilNotificationTime()
    {
        long nowMillis = System.currentTimeMillis();
        long untilMillis = m_notifierTask.getNextScheduledDate().getTime();
                 
        long duration = untilMillis - nowMillis;
        
        if (duration <=0) 
            return;
            
        log.debug("Sleeping until: " + m_notifierTask.getNextScheduledDate());
        try
        {
            Thread.sleep( duration );
        }
        catch (InterruptedException e)
        {
            log.debug( "Interrupted!  Is it time or too early?" );
            //ignore...
        }

    }


    /**
     * Compares system clock to tasks next scheduled time, returns true if it is time or 
     * past time for the task to be executed.
     */
    private boolean isTimeToNotify()
    {
        long nowMillis = System.currentTimeMillis();
        long untilMillis = m_notifierTask.getNextScheduledDate().getTime();

        return (untilMillis <= nowMillis);
    }
    
    
    

}
