/*
 * Created on Dec 15, 2003
 */
package com.mckessonaps.jspwiki.plugin.query;

import java.io.FileInputStream;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Properties;
import java.util.TreeSet;

import junit.framework.TestCase;
import junit.textui.TestRunner;

import com.ecyrd.jspwiki.ReferenceManager;
import com.ecyrd.jspwiki.ReferenceManagerSpy;
import com.ecyrd.jspwiki.TestEngine;
import com.ecyrd.jspwiki.WikiEngine;

/**
 * @author jvolkar
 */
public class QueryPluginTest extends TestCase
{
    public static void main(String[] args)
    {
        TestRunner.run(QueryPluginTest.class);
    }
    
    WikiEngine m_engine;
    ReferenceManager m_referenceManager;
    private QueryPlugin m_plugin;
    private Map m_params;
    

    public void testExecuteNoPageParameter()
    {
        assertEquals( "Missing Parameter: 'pages'", m_plugin.execute( m_engine, m_params));

        m_params.put( "pages", null );
        assertEquals( "Missing Parameter: 'pages'", m_plugin.execute( m_engine, m_params));
    }

    
    public void testBracketParsing()
    {
        m_params.put( "pages", "foo [some [more] bar" );
        assertEquals( "[]'s are nested.", m_plugin.execute( m_engine, m_params));
        
        m_params.put( "pages", "foo [some more]] bar" );
        assertEquals( "[]'s are mismatched.", m_plugin.execute( m_engine, m_params));
        
        m_params.put( "pages", "foo [some more][ bar" );
        assertEquals( "[]'s are not balanced.", m_plugin.execute( m_engine, m_params));

        m_params.put( "pages", "Anything without brackets." );
        assertEquals( "Page names must be surrounded by []'s.", m_plugin.execute( m_engine, m_params));            
    }

    
    
    public void testParenthesisParsing()
    {
        
        m_params.put( "pages", "foo ([page] more)) bar" );
        assertEquals( "()'s are mismatched.", m_plugin.execute( m_engine, m_params));
        
        m_params.put( "pages", "foo ([page] more)( bar" );
        assertEquals( "()'s are not balanced.", m_plugin.execute( m_engine, m_params));

    }
    
    public void testExpressionParsing()
    {
        final String STANDARD_FORM_TO_SPOKE_B = "(TO ([SpokeB]))";

        
        checkParsing(STANDARD_FORM_TO_SPOKE_B, "TO [SpokeB]");
        checkParsing(STANDARD_FORM_TO_SPOKE_B, "(TO [SpokeB])" );
        checkParsing(STANDARD_FORM_TO_SPOKE_B, "TO ([SpokeB])" );
        checkParsing(STANDARD_FORM_TO_SPOKE_B, "(TO ([SpokeB]))" );
        checkParsing(STANDARD_FORM_TO_SPOKE_B, "((TO ([SpokeB])))" );
        checkParsing(STANDARD_FORM_TO_SPOKE_B, "((TO (([SpokeB]))))" );
        
        
        //Okay try a vague complicated one...
        String expectedForm =      "([SpokeB] OR ([BranchB] AND ([Island] AND (NOT ([BranchB])))))";
        checkParsing(expectedForm, "( [SpokeB] OR [BranchB] AND [Island] NOT [BranchB] )" );
        
    }

    
    private void checkParsing(final String standardForm, String userForm)
    {
        m_params.put( "pages", userForm );
        m_plugin.execute( m_engine, m_params);
        String parsedForm = (String)m_params.get( "pages" );
        assertEquals(standardForm, parsedForm);
    }


    public void testExpressionEvaluation()
    {
        String expected;
        
        //Implied IS modifier...
        expected = "SpokeB";
        checkEvaluation(expected, "[SpokeB]");
        
        //Explicit IS...
        expected = "SpokeB";
        checkEvaluation(expected, "IS [SpokeB]");
        
        //Simple OR and IS...
        expected = "SpokeB,SpokeA";
        checkEvaluation(expected, "[SpokeB] OR [SpokeA]" );
        checkEvaluation(expected, "IS [SpokeB] OR [SpokeA]" );
        checkEvaluation(expected, "IS ([SpokeB] OR [SpokeA])" );
        checkEvaluation(expected, "IS [SpokeB] OR IS [SpokeA]" );
        
        //Simple TO...
        expected = "Hub";
        checkEvaluation(expected, "TO [SpokeB]" );
        checkEvaluation(expected, "TO ([SpokeB])" );
        checkEvaluation(expected, "(TO [SpokeB])" );
        checkEvaluation(expected, "(TO ([SpokeB]))" );
        
        //Basic TO and OR...
        expected = "Hub,Tree";
        checkEvaluation(expected, "TO [SpokeB] OR [BranchB]" );
        checkEvaluation(expected, "(TO [SpokeB] OR [BranchB])" );
        checkEvaluation(expected, "TO ([SpokeB] OR [BranchB])" );
        checkEvaluation(expected, "(TO ([SpokeB] OR [BranchB]))" );
        
        //Basic TO and AND...
        expected = "Hub";
        checkEvaluation(expected, "TO [SpokeB] AND [SpokeC]" );
        checkEvaluation(expected, "(TO [SpokeB] AND [SpokeC])" );
        checkEvaluation(expected, "TO ([SpokeB] AND [SpokeC])" );
        checkEvaluation(expected, "(TO ([SpokeB] AND [SpokeC]))" );
        
        
        //Simple BY...
        expected = "Rim";
        checkEvaluation(expected, "BY [SpokeB]" );
        checkEvaluation(expected, "BY ([SpokeB])" );
        checkEvaluation(expected, "(BY [SpokeB])" );
        checkEvaluation(expected, "(BY ([SpokeB]))" );
        
        //Basic BY and OR...
        expected = "Rim,BranchB1,BranchB2,BranchB3";
        checkEvaluation(expected, "BY [SpokeB] OR [BranchB]" );
        checkEvaluation(expected, "(BY [SpokeB] OR [BranchB])" );
        checkEvaluation(expected, "BY ([SpokeB] OR [BranchB])" );
        checkEvaluation(expected, "(BY ([SpokeB] OR [BranchB]))" );
        
        //Basic BY and AND...
        expected = "Rim";
        checkEvaluation(expected, "BY [SpokeA] AND [Common]" );
        checkEvaluation(expected, "(BY [SpokeA] AND [Common])" );
        checkEvaluation(expected, "BY ([SpokeA] AND [Common])" );
        checkEvaluation(expected, "(BY ([SpokeA] AND [Common]))" );
        

        //Hmm, Try NOT...
        expected = "Island,Common,Hub,Tree,Rim,BranchB1,BranchB2,BranchB3";
        checkEvaluation(expected, "NOT BY [Hub] OR [Tree]" );
        checkEvaluation(expected, "NOT BY ([Hub] OR [Tree])" );
        checkEvaluation(expected, "(NOT BY [Hub] OR [Tree])" );
        checkEvaluation(expected, "NOT (BY [Hub] OR [Tree])" );
        checkEvaluation(expected, "NOT (BY ([Hub] OR [Tree]))" );
        checkEvaluation(expected, "(NOT (BY ([Hub] OR [Tree])))" );
        

        //Okay so far BY, TO, IS, AND, OR, NOT all work single or in combination...
        
        
        //Try a tangled query without getting all ()'s on them, how well does it 
        //sound and work?
        expected = "Rim,Island";
        //read this: "BY [Common] AND NOT TO [SpokeB] AND NOT TO [BranchA] OR IS [Island]" 
        //as an english sentence after drawing out the setup pages...
        //it does not parse as I'd like it without any paren's, but that's OKAY, as
        //requireing parens for clarity is good, a simplistic paren'ing to show intent is...
        checkEvaluation(expected, "(  (BY [Common]) AND (NOT TO [SpokeB]) AND (NOT TO [BranchA])  ) OR ( IS [Island] )" );
        
        //Parsed formal form is:
        // (((BY ([Common])) AND ((NOT (TO ([SpokeB]))) AND (NOT (TO ([BranchA]))))) OR (IS ([Island]))
        
    }

    
    public void testLaterFoundBugs()
    {
        String expected;
        
        //Suspected during testing of the Visualizer, but apparently NOT a bug!
        expected = "Hub,Rim";
        checkEvaluation(expected, "TO [SpokeA] OR BY [SpokeA]" );
        checkEvaluation(expected, "(TO [SpokeA]) OR (BY [SpokeA])" );
        checkEvaluation(expected, "((TO [SpokeA]) OR (BY [SpokeA]))" );
    }
    
    
    public void testTermVariations()
    {
        String expected;
        
        //Implied IS modifier, testing spaces, and plurals...
        expected = "SpokeB";
        checkEvaluation(expected, "[SpokeB]");
        //fails checkEvaluation(expected, "[Spoke B]");
        //fails checkEvaluation(expected, "[SpokeBs]");
        //fails checkEvaluation(expected, "[Spoke Bs]");
    }    
    

    private void checkEvaluation(final String expectedPages, String userQuery)
    {
        m_params.put( "pages", userQuery );
        
        String resultsHtml = m_plugin.execute( m_engine, m_params);
        
        //Just to look at it...
        String parsedForm = (String)m_params.get( "pages" );
        
        String expectedHtml = m_plugin.processResultSet(m_engine, new TreeSet( Arrays.asList( expectedPages.split( "," ) ) ) );
        
        assertEquals(expectedHtml, resultsHtml);
        
        
    }
    
    
    
    /* (non-Javadoc)
     * @see junit.framework.TestCase#setUp()
     */
    protected void setUp() throws Exception
    {
        Properties props = new Properties();
        props.load(  new FileInputStream("test-jspwiki.properties" ));
        
        m_engine =  new TestEngine(props);
        m_referenceManager = m_engine.getReferenceManager();
        
        setupTestPages();
        
        m_plugin = new QueryPlugin();
        m_params = new HashMap();

    }

    
    
    /* (non-Javadoc)
     * @see junit.framework.TestCase#tearDown()
     */
    protected void tearDown() throws Exception
    {
    }

    /**
     * Test data in the reference manager looks like...
     * 
     *   Hub -> Spoke(A, B, C)
     *   Spoke(A, B, C) -> Rim
     * 
     *   Tree -> Branch(A, B, C)
     *   BranchB -> BranchB(1, 2, 3)
     * 
     *   Common -> (Hub, Tree)
     * 
     *   Island
     * 
     * 
     * NOTE: When constucting the test mesh, be careful of aliasing of the Lists
     * passed into the reference manager!!!
     */
    private void setupTestPages()
    {
        //Add Hub pointing to Spokes ABC
        List spokeList = new ArrayList();
        spokeList.add( "SpokeA");
        spokeList.add( "SpokeB");
        spokeList.add( "SpokeC");
        m_referenceManager.updateReferences( "Hub", spokeList );
        
        //Add 'Rim' pointing to no-one...
        List tempList = new ArrayList();
        m_referenceManager.updateReferences( "Rim", tempList );
        
        //Add spokes, each pointing to Rim...
        tempList = new ArrayList();
        tempList.add( "Rim");
        for (Iterator iter = spokeList.iterator(); iter.hasNext();)
        {    
            m_referenceManager.updateReferences( (String)iter.next(), tempList);
        }

        //Add Tree pointing to Branches
        List branchList = new ArrayList();
        branchList.add( "BranchA");
        branchList.add( "BranchB");
        branchList.add( "BranchC");
        m_referenceManager.updateReferences( "Tree", branchList );

        //Add Branches ABC pointing to no-one
        tempList= new ArrayList();
        for (Iterator iter = branchList.iterator(); iter.hasNext();)
        {    
            m_referenceManager.updateReferences( (String)iter.next(), tempList);
        }
        
        
        //Add BranchB pointing to BranchB 123...
        branchList= new ArrayList();
        branchList.add( "BranchB1");
        branchList.add( "BranchB2");
        branchList.add( "BranchB3");
        m_referenceManager.updateReferences( "BranchB", branchList);
        
        //Add BranchB 123 pointing to no-one...
        tempList= new ArrayList();
        for (Iterator iter = branchList.iterator(); iter.hasNext();)
        {    
            m_referenceManager.updateReferences( (String)iter.next(), tempList);
        }
        
        //Add Common pointing to Hub and Tree
        tempList= new ArrayList();
        tempList.add( "Hub");
        tempList.add( "Tree");
        tempList.add( "Rim");
        m_referenceManager.updateReferences( "Common", tempList);
        
        //Add Island pointing to no-one...
        tempList = new ArrayList();
        m_referenceManager.updateReferences( "Island", tempList);
        debugMaps();
        
        
    }
    
    private void debugMaps()
    {
        ReferenceManagerSpy spy = new ReferenceManagerSpy(m_engine);
        
        Map toMap = spy.getRefersTo();
        Map byMap = spy.getReferredBy();

    }
}
