package com.mckessonaps.jspwiki.plugin.redirect;


import java.util.Map;

import org.apache.log4j.Logger;

import com.ecyrd.jspwiki.TextUtil;
import com.ecyrd.jspwiki.WikiContext;
import com.ecyrd.jspwiki.plugin.WikiPlugin;
import com.mckessonaps.jspwiki.plugin.util.ParamUtil;

/**
 * Will return a fragment of javascript to do an automatic redirection.<br/>
 * Requires only one parameter 'to', which can be InterWiki link, if it cannot be 
 * interperated as an interwiki we assume it's some external link and just redirect 
 * to it as given.  If it <b>is</b> an interwiki, we first expand it then redirect.
 * <p>  
 * Warning:  If you use this plugin it can be hard to actually edit the page that contains 
 * the redirect plugin instance, you must hand edit a URL or have an edit link elsewhere.
 * <p> 
 * Usage:
 * <pre>
 * [{INSERT RedirectPlugin to='http://www.google.com' delay='0'}]
 * 
 * or
 * 
 * [{INSERT RedirectPlugin to='JSPWiki:Main'}]
 * </pre>
 */
public class RedirectPlugin 
implements WikiPlugin
{
    private static final Logger log = Logger.getLogger( RedirectPlugin.class );


    /**
     * Execute this plugin, look for the 'to' parameter, return fragment of javascript to do the redirection.
     * @see com.ecyrd.jspwiki.plugin.WikiPlugin#execute(com.ecyrd.jspwiki.WikiContext, java.util.Map)
     */
    public String execute(WikiContext context, Map paramMap) 
    {
        //Get the one required parameter...
        String toTarget = (String)paramMap.get( "to" );
        
        //Get the optional delay parameter
        int delayMs = ParamUtil.getInteger( log, paramMap, "delay", 2000 );
        String delay = Integer.toString( delayMs );
        
        //'to' Cannot be null...
        if (null == toTarget)
        {
            log.info( "User forgot the 'to' parameter." );
            return "error! parameter 'to' is required.";
        }
        
        //'to' Must not be empty....
        if (toTarget.trim().equals("") )
        {
            log.info( "User gave a blank 'to' parameter." );
            return "error! parameter 'to' must not be blank.";
        }
        
        //What's the 'to' look like, interwiki?
        int colonPosition = toTarget.indexOf(':');
        
        if (-1 == colonPosition)
        {
            return redirectTo(toTarget, context.getEngine().getViewURL( toTarget ), delay); 
        }
            
        String targetProtocol = toTarget.substring( 0, colonPosition);
        
        //Okay, lets try to interperate it as an interwiki link....
        String interwikiUrlPattern = context.getEngine().getInterWikiURL( targetProtocol );
        
        //If we could not find an interwiki link defined, just redirect to it raw...
        if (null == interwikiUrlPattern)
            return redirectTo(toTarget, toTarget, delay);
            
        //Extract the right-side of the original 'to' value...
        String interwikiTarget = toTarget.substring( colonPosition + 1 );
        String expandedInterwikiUrl = TextUtil.replaceString( interwikiUrlPattern, "%s", interwikiTarget );
        
        return redirectTo(toTarget, expandedInterwikiUrl, delay);
    }
     
     
     
     /**
      * Form the javascript fragment to do redirection...
      * @param toUrl
      * @return String
      */
     private String redirectTo(String toTarget, String toUrl, String delay)   
    {
        String redirectScript = "";
        
        redirectScript +="<b>Redirecting to: <i>";
        redirectScript +="<a href='" + toUrl + "'>" + toTarget + "</a>";
        redirectScript +="</i></b>";
         
        redirectScript +="<script language='javascript'>\n";
        redirectScript +="<!--\n";
        redirectScript +="function doReplace(){location.replace('" + toUrl + "');}\n";
        redirectScript +="setTimeout('doReplace()', "+ delay +")\n";
        redirectScript += "-->\n";
        redirectScript += "</script>\n";
        
        return redirectScript;
    }

}
