package com.mckessonaps.jspwiki.plugin.ticker;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.net.MalformedURLException;
import java.net.URL;
import java.net.URLConnection;
import java.util.HashMap;
import java.util.Map;

import org.apache.log4j.Category;

import com.ecyrd.jspwiki.WikiContext;
import com.ecyrd.jspwiki.plugin.PluginException;
import com.ecyrd.jspwiki.plugin.WikiPlugin;

/**
 * The following exchanges are supported by Yahoo finance.
 * <a href='http://finance.yahoo.com/sd'>http://finance.yahoo.com/sd</a>
 */
public class TickerPlugin
implements WikiPlugin
{
    private static final Category log = Category.getInstance( TickerPlugin.class );

    private static final String CSV_FORMAT ="l1";//Long format: "sl1d1t1c1ohgv";
    
    private static final Map c_cachedQuotesMap = new HashMap();


    
//If needed... 
//import java.net.Authenticator;
//import java.net.PasswordAuthentication;
//    {
//        Authenticator.setDefault( new Authenticator()
//        {
//            PasswordAuthentication m_authentication = new PasswordAuthentication("userid", "password".toCharArray() );  
//
//            protected PasswordAuthentication getPasswordAuthentication()
//            {
//                return m_authentication;
//            }
//        });
//    }


    
    /* (non-Javadoc)
     * @see com.ecyrd.jspwiki.plugin.WikiPlugin#execute(com.ecyrd.jspwiki.WikiContext, java.util.Map)
     */
    public String execute(WikiContext context, Map paramMap) 
    throws PluginException
    {
        String symbol = (String)paramMap.get("symbol");
        if (null == symbol) throw new PluginException("Must provide 'symbol' parameter.");
        symbol = symbol.trim();
        
        String link = "http://finance.yahoo.com/q?s=" + symbol + "&d=t";
        
        return symbol + ":<a href='" + link + "'>" + lookupValue(symbol) + "</a>";
    }



    /**
     * get quote for symbol
     */
    private String lookupValue(String symbol)
    {
        Quote q = (Quote)c_cachedQuotesMap.get( symbol );
        
        if (null == q) 
        {
            q = new Quote( symbol, fetchValue(symbol));
            c_cachedQuotesMap.put( q.getSymbol(), q);
            log.debug( "Adding quoute for '"+ q.getSymbol() +"' to cache.");
        }
                
        if (q.isExpired()) q.update( fetchValue(symbol) );
        
        return q.getValue();
    }
    
    /**
     * Return the current (time-lagged-yahoo) value of symbol. 
     * @param symbol
     * @return String
     */
    private String fetchValue(String symbol)
    {
        URL url = null;
        URLConnection conn = null;
        Object content = null;
        String valueString = "error!";
        try
        {
            url = makeUrl(symbol);
            conn = url.openConnection();
            content = conn.getContent();            
            valueString = extractValue(content);
        }
        catch (MalformedURLException e)
        {
            log.warn( "Ignoring exception.", e);
        }
        catch (IOException e)
        {
            log.warn( "Ignoring exception.", e);
        }
        
        return valueString;
    }
    

    /**
     * Make yahoo finance url...
     * @param symbol
     * @return URL
     * @throws MalformedURLException
     */
    private URL makeUrl(String symbol)
    throws MalformedURLException
    {
        return new URL( "http://finance.yahoo.com/d/quotes.csv?e=.csv&f=" + CSV_FORMAT + "&s=" + symbol);
    }
    

    /**
     * Get quote value from Object.
     * @param o
     * @return String
     */
    private String extractValue(Object o)
    {
        if (o instanceof String) 
            return extractValue((String)o);
        else if (o instanceof InputStream) 
            return extractValue((InputStream)o);
        else
            return "error!";
           
    }
    
    
    /**
     *  csvLine could look smething like...
     * <pre>
     *  ticker symbol, last price, date, time, change, open price, daily high, daily low, and volume. 
     *  "AAPL",22.25,"10/16/2000","1:02PM",+0.1875,22.3125,23.25,22.0625,8263600
     *   s      l1    d1           t1       c1      o       h     g       v
     * </pre>
     * Depends on how much data we want brought back, see CSV_FORMAT, currently
     * we only ask for 'l1', so nothing to do here but maybe someday we'll ask for more?
     * 
     * @param csvLine
     * @return String
     */
    private String extractValue(String csvLine)
    {
        return csvLine;
    }
    
    
    /**
     * Get quote value from stream.
     * @param is
     * @return String
     */
    private String extractValue(InputStream is)
    {
        InputStreamReader isr = null; 
        BufferedReader br = null;
        String valueString = "error!";
        try
        {
            isr = new InputStreamReader(is);
            br = new BufferedReader(isr);
            valueString = extractValue( br.readLine() );
        }
        catch(IOException ioe)
        {
            //ignore
            log.warn( "Ignoring exception.", ioe);
        }
        finally
        {
            if (null != br) try { br.close(); } catch (IOException ioe){}
            if (null != isr) try { isr.close(); } catch (IOException ioe){}
            if (null != is) try { is.close(); } catch (IOException ioe){}
        }
        
        return valueString;
    }




    /**
     * Silly little testing method.
     */
    public static void main(String[] args)
    {
        TickerPlugin ticker = new TickerPlugin();
        Map map = new HashMap();
        try
        {
            map.put( "symbol", "IBM");
            System.out.println( ticker.execute( null, map) );
                    
            map.put( "symbol", "IBM.BA");
            System.out.println( ticker.execute( null, map) );
            
            
            //now cache hits...
            
            
            map.put( "symbol", "IBM");
            System.out.println( ticker.execute( null, map) );
                    
            map.put( "symbol", "IBM.BA");
            System.out.println( ticker.execute( null, map) );
        }
        catch (Exception e)
        {
            e.printStackTrace();
        }
        
    }
    
    
}
